Require Import ssreflect ssrbool eqtype seq fintype bigop ssralg ssrnum.

(******************************************************************************)
(*                                                                            *)
(* Here is define the maximum and the minimum for many elementsand we proof   *)
(* some basic result about them.                                              *)
(*                                                                            *)
(*   \maxr_(i <- s) F i == \big[Num.max/0]_(i <- s) F i. It is the maximum    *)
(*                         of all F i only if each F i are non-negative.      *)
(*                         (since the neutral elements is 0).                 *)
(*                         (see the file bigop.v  of SSReflect library for    *)
(*                          this notation)                                    *)
(*                                                                            *)
(*                                                                            *)
(*            min_seq s == it is the minimum of elements of the sequence s.   *)
(*                         that return 0 if s is nil.                         *)
(*                                                                            *)
(*             min_fT F == if F : fT -> rT where fT is a finType then that    *)
(*                         return the minimum of all F i forall i of Type fT  *)
(*                         and 0 if fT is not inhabited.                      *)
(******************************************************************************)


Import GRing.Theory.
Import Num.Theory.
Import Num.Def. 

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Reserved Notation "\maxr_ i F"
  (at level 41, F at level 41, i at level 0,
           format "'[' \maxr_ i '/ '  F ']'").
Reserved Notation "\maxr_ ( i <- r | P ) F"
  (at level 41, F at level 41, i, r at level 50,
           format "'[' \maxr_ ( i  <-  r  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i <- r ) F"
  (at level 41, F at level 41, i, r at level 50,
           format "'[' \maxr_ ( i  <-  r ) '/ '  F ']'").
Reserved Notation "\maxr_ ( m <= i < n | P ) F"
  (at level 41, F at level 41, i, m, n at level 50,
           format "'[' \maxr_ ( m  <=  i  <  n  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( m <= i < n ) F"
  (at level 41, F at level 41, i, m, n at level 50,
           format "'[' \maxr_ ( m  <=  i  <  n ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i | P ) F"
  (at level 41, F at level 41, i at level 50,
           format "'[' \maxr_ ( i  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i : t | P ) F"
  (at level 41, F at level 41, i at level 50,
           only parsing).
Reserved Notation "\maxr_ ( i : t ) F"
  (at level 41, F at level 41, i at level 50,
           only parsing).
Reserved Notation "\maxr_ ( i < n | P ) F"
  (at level 41, F at level 41, i, n at level 50,
           format "'[' \maxr_ ( i  <  n  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i < n ) F"
  (at level 41, F at level 41, i, n at level 50,
           format "'[' \maxr_ ( i  <  n )  F ']'").
Reserved Notation "\maxr_ ( i 'in' A | P ) F"
  (at level 41, F at level 41, i, A at level 50,
           format "'[' \maxr_ ( i  'in'  A  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i 'in' A ) F"
  (at level 41, F at level 41, i, A at level 50,
           format "'[' \maxr_ ( i  'in'  A ) '/ '  F ']'").


Section maxr.

Variable T : numDomainType.

Open Scope ring_scope.

Lemma max0r (x : T) :  (0 <= x) -> (maxr 0 x = x) .
Proof.
move=> H; rewrite /maxr; case: ifP=> //.
by rewrite ler_eqVlt real_ltrNge ?real0 ?ger0_real // H orbF=> /eqP ->.
Qed.

Lemma real_ler_maxr (x y z : T) : 
  x \is Num.real ->  y \is Num.real ->  z \is Num.real -> 
  (x <= maxr y z) = (x <= y) || (x <= z). 
Proof.
move=> rx ry rz; rewrite /maxr; case: ifP; case: (real_lerP rx rz)=> xlz zly.
   by rewrite (ler_trans xlz zly).
   by rewrite orbF.
   by rewrite orbT.
   by rewrite (real_lerNgt rx ry) (ltr_trans _ xlz) // (real_ltrNge ry rz) zly.
Qed.

Lemma real_ltr_maxr (x y z : T) : 
  x \is Num.real ->  y \is Num.real ->  z \is Num.real -> 
  (x < maxr y z) = (x < y) || (x < z). 
Proof.
move=> rx ry rz; rewrite /maxr; case: ifP; case: (real_ltrP rx rz)=> xlz zly.
   by rewrite (ltr_le_trans xlz zly).
   by rewrite orbF.
   by rewrite orbT.
rewrite (real_ltrNge rx ry) (ler_trans _ xlz) // ltrW //.
by rewrite (real_ltrNge ry rz) zly.
Qed.

Lemma real_ler_maxl (x y z : T) :
    x \is Num.real ->  y \is Num.real ->  z \is Num.real -> 
   (maxr x y <= z) = (x <= z) && (y <= z).
Proof.
move=> rx ry rz; rewrite /maxr; case: ifP; case: (real_lerP rx rz)=> xlz ylx //.
  by rewrite (ler_trans ylx xlz).
by rewrite (real_lerNgt ry rz) (ltr_trans xlz) // (real_ltrNge rx ry) ylx.
Qed.

Lemma real_ltr_maxl (x y z : T) :
    x \is Num.real ->  y \is Num.real ->  z \is Num.real -> 
   (maxr x y < z) = (x < z) && (y < z).
Proof.
move=> rx ry rz; rewrite /maxr; case: ifP; case: (real_ltrP rx rz)=> xlz ylx //.
  by rewrite (ler_lt_trans ylx xlz).
by rewrite (real_ltrNge ry rz) (ler_trans xlz) ?ltrW ?(real_ltrNge rx ry) ?ylx.
Qed.

End maxr.



Notation "\maxr_ ( i <- r | P ) F" :=
  (\big[Num.max/0%R]_(i <- r | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i <- r ) F" :=
  (\big[Num.max/0%R]_(i <- r) F%R) : ring_scope.
Notation "\maxr_ ( i | P ) F" :=
  (\big[Num.max/0%R]_(i | P%B) F%R) : ring_scope.
Notation "\maxr_ i F" :=
  (\big[Num.max/0%R]_i F%R) : ring_scope.
Notation "\maxr_ ( i : I | P ) F" :=
  (\big[Num.max/0%R]_(i : I | P%B) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( i : I ) F" :=
  (\big[Num.max/0%R]_(i : I) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( m <= i < n | P ) F" :=
 (\big[Num.max/0%R]_(m <= i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( m <= i < n ) F" :=
 (\big[Num.max/0%R]_(m <= i < n) F%R) : ring_scope.
Notation "\maxr_ ( i < n | P ) F" :=
 (\big[Num.max/0%R]_(i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i < n ) F" :=
 (\big[Num.max/0%R]_(i < n) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A | P ) F" :=
 (\big[Num.max/0%R]_(i in A | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A ) F" :=
 (\big[Num.max/0%R]_(i in A) F%R) : ring_scope.

Section bigmax.

Variable R : numDomainType.
Open Scope ring_scope.

Lemma bigmax_pos (T: numDomainType) (I : Type) (F : I -> T) (s : seq I) :
  (forall i, 0 <= F i) -> 0 <= \maxr_(i <- s) F i.
Proof.
move=> Hge0.
apply: (big_ind (fun x : T => 0 <= x))=> // x y Hx Hy.
by rewrite /maxr; case: (y <= x).
Qed.

Lemma bigmax_ger (T: numDomainType) (I : eqType) (F : I -> T) (s : seq I) i :
(forall i, 0 <= F i) ->  i \in s -> F i <= \maxr_(i <- s) F i.
Proof.
move=> Hge0.
elim: s=> // a l IHl.
set x := F a; set y := \big[_/_]_(_ <- _) _.
have Hxr := ger0_real (Hge0 a).
have Hyr := ger0_real (bigmax_pos l Hge0).
rewrite in_cons /y big_cons /maxr=> /orP []=> [/eqP ->|/IHl Hi];
case: (real_lerP Hyr Hxr)=> //.
  by move/ltrW.
exact: ler_trans.
Qed.

Lemma bigmax_eq (T : numDomainType) (I : eqType) (F : I -> T) (s : seq I) :
  (forall i, 0 <= F i) ->  s != [::] -> 
  {k | (k \in s) && (F k == \maxr_(i <- s) F i)}.
Proof.
move=> Hge0.
elim: s=> // a []=> [_ _|b l /(_ is_true_true) [k /andP [Hks Hk]] _].
  by exists a; rewrite big_cons big_nil /maxr Hge0 mem_seq1 !eqxx.
rewrite big_cons /maxr.
set y := \big[_/_]_(_ <- _) _.
case: (y <= F a); [exists a| exists k].
  by rewrite mem_head eqxx.
by rewrite Hk mem_behead // Hks.
Qed.

Lemma bigmax_def (I : eqType) (F : I -> R) (s : seq I) x :
  (forall i, 0 <= F i) -> {i | (x == F i) && (i \in s) } ->
  (forall j, j \in s -> F j <= x) ->  
  \maxr_(k <- s) F k = x.
Proof. 
move=> Hge0 [] i /andP [] /eqP ->.
elim: s=> // a []=> [_|b l IHl].
  by rewrite mem_seq1 big_cons big_nil /maxr Hge0=> /eqP ->.
rewrite in_cons big_cons.
have [k /andP [Hks /eqP Hk]] := bigmax_eq (s := b :: l) Hge0 is_true_true.
case/orP=> [/eqP Hi|Hi] Hge; rewrite /maxr -Hk.
  by rewrite -Hi Hge // mem_behead.
have := Hge a; rewrite -IHl // =>[|j Hj].
  rewrite mem_head -Hk ler_eqVlt=> /(_ is_true_true)/orP []=> [/eqP ->|].
    by rewrite lerr.
  by rewrite (real_lerNgt (ger0_real (Hge0 k)) (ger0_real (Hge0 a)))=> ->.
by rewrite Hge // mem_behead.
Qed.

Lemma bigmax_distrl (I : eqType) (F : I -> R) (s : seq I) (r : R) :
  0 <= r -> (r * \maxr_(i <- s) (F i)) = \maxr_(i <- s) (r * F i).
Proof.
rewrite ler_eqVlt=> le0r.
elim: s=> [|a l IHl]; first by rewrite !big_nil mulr0.
rewrite !big_cons -IHl /maxr.
set x := F a; set y := \big[_/_]_(_ <- _) _.
case/orP: le0r=> [/eqP <-|lt0r]; first by rewrite !mul0r lerr.
by rewrite ler_pmul2l //; case: (y <= x).
Qed.

Lemma bigmax_ler_seq (I : eqType) (F1 F2 : I -> R) (s : seq I) :
 (forall i, 0 <= F1 i) ->  (forall i, i\in s -> F1 i <= F2 i) -> 
 \maxr_(i <- s) F1 i <= \maxr_(i <- s) F2 i.
Proof.
move=> Hge0.
elim: s=> [|a l IHl Hle]; first by rewrite !big_nil.
rewrite !big_cons /maxr.
have IHle i: i \in l -> F1 i <= F2 i by move=> H; rewrite Hle // mem_behead.  
have x1r := ger0_real (Hge0 a).
have x2r := ger0_real (ler_trans (Hge0 a) (Hle a (mem_head a l))).
have y1r := ger0_real (bigmax_pos l Hge0).
have y2r := ger0_real (ler_trans (bigmax_pos l Hge0) (IHl IHle)). 
case: (real_lerP y1r x1r)=>[_|_]; case: (real_lerP y2r x2r)=>[Hxy|Hyx].
  by rewrite Hle // mem_head.
  by rewrite ltrW // (ler_lt_trans (Hle a (mem_head a l)) Hyx).
  by rewrite (ler_trans (IHl IHle) Hxy).
exact: IHl.
Qed.

Lemma bigmax_ler (I : finType) (F1 F2 : I -> R) :
 (forall i, 0 <= F1 i) ->  (forall i,  F1 i <= F2 i) -> 
 \maxr_i F1 i <= \maxr_i F2 i.
Proof.
move=> Hge0 Hle.
elim: (index_enum I) => [|a l IHl]; first by rewrite !big_nil.
rewrite !big_cons /maxr.
have x1r := ger0_real (Hge0 a).
have x2r := ger0_real (ler_trans (Hge0 a) (Hle a)).
have y1r := ger0_real (bigmax_pos l Hge0).
have y2r := ger0_real (ler_trans (bigmax_pos l Hge0) IHl). 
case: (real_lerP y1r x1r)=>[_|_]; case: (real_lerP y2r x2r)=>[Hxy|Hyx] //.
  by rewrite ltrW // (ler_lt_trans (Hle a) Hyx).
by rewrite (ler_trans IHl Hxy).
Qed.
 
Lemma bigmax_ler_sum (I : finType) (F1 F2 : I -> R) :
  (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) ->
  \maxr_i (F1 i + F2 i) <=  (\maxr_i F1 i) + \maxr_i F2 i.
Proof.
move=> H1ge0 H2ge0.
elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil addr0.
rewrite !big_cons /maxr.
set x1 := F1 a; set y1 := \big[_/_]_(_ <- _) _.
set x2 := F2 a; set y2 := \big[_/_]_(_ <- _) _.
have x1r := ger0_real (H1ge0 a).
have x2r := ger0_real (H2ge0 a).
have y2r := ger0_real (bigmax_pos l H1ge0).
have zr := ger0_real (bigmax_pos l H2ge0).
have y1r : y1 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i; rewrite addr_ge0.
have x12r : x1 + x2 \is Num.real by rewrite realD.
case: (real_lerP y1r x12r)=> [_|_]; case: (real_lerP y2r x1r)=>[c|d]; 
case: (real_lerP zr x2r)=> [g|h] //.
  +by rewrite ler_add2l ltrW.
  -by rewrite ler_add2r ltrW.
  +by rewrite ltrW // ltr_add.
  -by rewrite (ler_trans IHl) // ler_add.
  +by rewrite (ler_trans IHl) // ler_add2r.
  -by rewrite (ler_trans IHl) // ler_add2l.
Qed.

Lemma bigmax_le_prod2 (I J : finType) (F1 : I -> R) (F2 : J -> R) :
  (forall i, 0 <= F1 i) -> (forall j, 0 <= F2 j) ->
  \maxr_i \maxr_j (F1 i * F2 j) <= (\maxr_i F1 i) * \maxr_j F2 j.
Proof.
move=> H1ge0 H2ge0.
elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil mul0r.
rewrite !big_cons -bigmax_distrl //.
set x := F1 a; set y1 := \big[_/_]_(_ <- _) F2 _.
set y2 := \big[_/_]_(_ <- _) \big[_/_]_ _ _; set z := \big[_/_]_(_ <- _) _.
have xr := ger0_real (H1ge0 a).
have zr : z \is Num.real := ger0_real (bigmax_pos l H1ge0).
have y1r : y1 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i.
have y2r : y2 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i; apply: bigmax_pos=> j; rewrite mulr_ge0.
have xy1r : x * y1 \is Num.real by rewrite realM.
rewrite /maxr.
case: (real_lerP y2r xy1r)=> [c'|d']; case: (real_lerP zr xr)=>[c|d] //. 
  by rewrite ler_pmul // ?H1ge0 ?bigmax_pos // ltrW.
by rewrite (ler_trans IHl) // ler_pmul // bigmax_pos.
Qed.

Lemma bigmax_le_prod (I : finType) (F1 F2 : I -> R) :
  (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) ->
  \maxr_i (F1 i * F2 i) <=  (\maxr_i F1 i) * \maxr_i F2 i.
Proof.
move=> H1ge0 H2ge0.
apply: ler_trans (bigmax_le_prod2 H1ge0 H2ge0).
have Hpos i j: 0 <= F1 i * F2 j by rewrite mulr_ge0.
by apply: bigmax_ler=> i //; apply: (bigmax_ger (Hpos i)).
Qed.

Lemma bigmax_eq0 (I : eqType) (F : I -> R) (s : seq I) :
  (forall i, i \in s -> F i = 0) -> \maxr_(i <- s) F i = 0.
Proof.
elim: s=>[|a l IHl Heq0]; first by rewrite big_nil.
rewrite big_cons IHl ?Heq0 /maxr ?lerr ?mem_head // => i Hi.
by rewrite Heq0 // mem_behead.
Qed.

Lemma map_bigmax (I : finType) (T1 : numDomainType) 
  (f : T1 -> R) (F : I -> T1) : 
  (forall i, 0 <= F i) -> (f 0 = 0) -> 
  (forall x y, x <= y -> f x <= f y) -> 
  f (\maxr_i F i) = \maxr_i f (F i).
Proof.
case: (index_enum I); first by rewrite !big_nil.
move=> a l Hpos Hf0 Hf; apply/sym_eq/bigmax_def.
  by move=> i; rewrite -Hf0 Hf.
have/(_ is_true_true) [k /andP[Hk HF]] := bigmax_eq Hpos (s:= a :: l). 
exists k; first by rewrite (eqP HF) eqxx.
by move=> j Hj; rewrite Hf // bigmax_ger.
Qed.

End bigmax.

Section bigmin.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Variable R : numDomainType.

Open Scope ring_scope.

Fixpoint min_seq (s : seq R) := 
  match s with 
   | [::] => 0
   | [:: x]  => x 
   | x :: s' => minr (min_seq s') x
  end.

Definition min_fT (I : finType) (F : I -> R) := min_seq (map F (index_enum I)).

Lemma min_seq_real (s : seq R) : (forall x, x \in s -> x \is Num.real) ->
 min_seq s \is Num.real.
Proof.
case: s=>[|a s]; first by rewrite real0.
elim: s a=>[a -> //|a l IHl b Hreal /=]; first by rewrite mem_head.
set c := (X in Num.min X _).
rewrite  /minr; case: (c <= b).
  by rewrite IHl // => x Hx; rewrite Hreal // mem_behead.
by rewrite Hreal // mem_head.
Qed.

Lemma min_seq_le (s : seq R) x : (forall x, x \in s -> x \is Num.real) -> 
  x \in s ->  min_seq s <= x.
Proof.
elim: s=> //= a []; first by rewrite in_cons orbF=> _ _ /eqP ->.
move=> b l IHl Hreal Hx.
have ar: a \is Num.real by rewrite Hreal // mem_head.
have mr: min_seq (b :: l) \is Num.real.
  by apply: min_seq_real=> y Hy; rewrite Hreal // mem_behead.
move: Hx.
rewrite in_cons /minr=>/orP []=>[/eqP ->|H]; case: (real_lerP mr ar)=> // [|Ha].
  by rewrite IHl // => y Hy; rewrite Hreal // mem_behead.
rewrite ltrW // (ltr_le_trans Ha) // IHl // => y Hy.
by rewrite Hreal // mem_behead.
Qed.

Lemma min_fT_le (I : finType) (F : I -> R) (i : I) :
 (forall i, F i \is Num.real) -> min_fT F <= F i.
Proof.
move=> Hreal; apply: min_seq_le=>[x /mapP [y Hy] -> //|].
by apply/mapP; exists i.
Qed.

Lemma ex_min_fT (I : finType) (F : I -> R) (a : I) : {k | min_fT F = F k}.
Proof.
rewrite /min_fT.
have := mem_index_enum a.
case: (index_enum I)=> // b l _. 
elim: l b=>[b|b l /= IHl c]; first by exists b.
have [k ->] := IHl b.
by rewrite /minr; case: (F k <= F c); [exists k|exists c].
Qed.

End bigmin.

